/*
 * MPC Feedback Module 4 for ATOM (instrumentation)
 *
 * Copyright (C) 2000-2004 by Grigori G. Fursin
 *
 * homepage: http://homepages.inf.ed.ac.uk/gfursin
 *
 * ICSA, School of Informatics,
 * University of Edinburgh, UK
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <machine/inst.h>
#include <cmplrs/atom.inst.h>

#define FuncMax 1024
#define LoopMax 1024
#define RegMax    32

void make_mpc_reg_iter(Inst*, int, int, int);

void InstrumentInit(int p1, char **p2) 
{
  AddCallProto("MPCInit(int, int, long)");
  AddCallProto("MPCFinish()");
  AddCallProto("MPCRegAccess(int)");
  AddCallProto("MPCCountProc(int)");
  AddCallProto("MPCSaveAddr(VALUE, int)");
  AddCallProto("MPCBigRegAccess(int, VALUE, long)");
  AddCallProto("MPCBigClean(int)");
  AddCallProto("MPCBigClean1(int)");
}

void fggstrlwr(char* addr)
{
  char c;

  while ((c=*addr)!=0)
  {
    *addr=tolower(c);
    addr++;
  }
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p;  
  Block *b; 
  Inst *i;

  FILE *fmpc1;
  FILE *fmpc2;
  FILE *fmpc5;

  char dummy[1024];
  char procname[1024];
  char procname1[1024];
  char addr1[16];
  char addr2[16];

  char fi[1024];

  char ops[1024];
  char fr[2];

  int line=0;
  int l1=0;
  int l2=0;

  int ret=0;
  int ret1=0;

  int ifound=0;
  int imd=0;
  int ira=0;
  int irb=0;
  int irc=0;
  int iop=0;
  int rw=0;

  int proc=0;
  int loop_max=0;
  int loop_max1=0;
  int inst_max=0;
  int inst_max1=0;

  int l_i4=0;
  int l_f8=0;

  int q1=0;
  int q2=0;

  int k=0;
  int j=0;

  int found1=0;

  int pass=0;
  
  long xaddr1=0;
  long xaddr2=0;

  /*************************** Main ***************************/
  printf("EOS MPC Feedback Module 4   (C) by Grigori G. Fursin   April,2002\n");

  pass=0;
  if (argc>1)
  {
    sscanf(argv[1], "%16ld", &pass);
  }

  strcpy(fi, "_eos_mpc_info2");
  if (argc>2)
  {
    strcpy(fi, argv[2]);
    strcat(fi, "2");
  }

  if (pass!=0)
  {
    if (pass>0)
      printf("\nSecond pass ...\n");
    else if (pass<0)
      printf("\nThird pass ...\n");

    if ((fmpc1=fopen("_eos_mpc_atom_tmp6b","rt"))==NULL)
    {
      fprintf(stderr, "\nFatal Error: Can't find file with parameters!\n");
      exit(1);
    }

    ret=fscanf(fmpc1, "%16ld", &xaddr1);
    ret=fscanf(fmpc1, "%16ld", &xaddr2);

    fclose(fmpc1);

    /* force to set xaddr1 to the following value */
    xaddr1=0x140000000;
    xaddr2+=16384;

    printf("\n");
    printf("Addr1: %16ld\n", xaddr1);
    printf("Addr2: %16ld\n", xaddr2);
  }

  if ((fmpc1=fopen(fi, "rt"))==NULL)
  {
    fprintf(stderr, "\nFatal Error: Can't find file with parameters!\n");
    exit(1);
  }

  if ((fmpc2=fopen("_eos_mpc_atom_tmp1_4","wt"))==NULL)
  {
    fprintf(stderr, "\nFatal Error: Can't open file for writing!\n");
    exit(1);
  }

  do
  {
    ret=fscanf(fmpc1, "%s", &dummy[0]);

    if (ret!=EOF && dummy[0]=='+')
    {
      ret=fscanf(fmpc1, "%s %s", &addr1[0], &addr2[0]);

      k=1;
      j=0;
      for (;k<strlen(dummy) && dummy[k]!=' '; k++) procname[j++]=dummy[k];
      procname[j]=0;

      fggstrlwr(procname);

      strcat(procname, "_");

      fprintf(fmpc2, "proc='%s', number='%u'\n", procname, proc);

      printf("\nSearching subroutine: %s (%s - %s)\n", procname, addr1, addr2);

      l1=atoi(addr1);
      l2=atoi(addr2);

      for (p=GetFirstObjProc(obj); p!=NULL; p=GetNextProc(p))
      {
        if (ProcName(p)!=NULL) strcpy(procname1, ProcName(p));
        else strcpy(procname1, "");

        fggstrlwr(procname1);

        if (strcmp(procname1, procname)==0)
        {   
          printf("Subroutine %s found!\n", procname);

          if (pass>=0)
            AddCallProc(p, ProcBefore, "MPCCountProc", proc);
          else if (pass<0)
            AddCallProc(p, ProcBefore, "MPCCountProc", 0);

          if (pass>0)
          {
            AddCallProc(p, ProcBefore, "MPCBigClean", proc);
          }
          else if (pass<0)
          {
            AddCallProc(p, ProcBefore, "MPCBigClean1", 0);
          }

          sprintf(dummy, "_eos_mpc_atom_tmp4_2.%u", proc);

          if ((fmpc5=fopen(dummy,"wt"))==NULL)
          {
            fprintf(stderr, "\nFatal Error: Can't open file for writing!\n");
            exit(1);
          }

          inst_max=0;

          for (b=GetFirstBlock(p); b!=NULL; b=GetNextBlock(b))
          {
            for (i=GetFirstInst(b); i!=NULL; i=GetNextInst(i))
            {
              line=InstLineNo(i);

              if ((l1<=line) && (line<=l2))
              {
                found1=0;

                if ((GetInstClass(i) == ClassLoad))
                {
                  found1=1;
                  strcpy(fr, "");
                }
                else if ((GetInstClass(i) == ClassFload))
                {
                  found1=2;
                  strcpy(fr, "f");
                }
                else if ((GetInstClass(i) == ClassStore))
                {
                  found1=3;
                  strcpy(fr, "");
                }
                else if ((GetInstClass(i) == ClassFstore))
                {
                  found1=4;
                  strcpy(fr, "f");
                }

                if (found1!=0)
                {
                  imd=GetInstInfo(i, InstMemDisp);
                  ira=GetInstInfo(i, InstRA);
                  irb=GetInstInfo(i, InstRB);
                  irc=GetInstInfo(i, InstRC);
                  iop=GetInstInfo(i, InstOpcode);

                  switch(iop)
                  {
                    case op_lda:   {strcpy(ops, "lda");     rw=0; break;};
                    case op_ldah:  {strcpy(ops, "ldah");    rw=0; break;};
                    case op_ldbu:  {strcpy(ops, "ldbu");    rw=0; break;};
                    case op_ldq_u: {strcpy(ops, "ldq_u");   rw=0; break;};
                    case op_ldwu:  {strcpy(ops, "ldwu");    rw=0; break;};
                    case op_stw:   {strcpy(ops, "stw");     rw=1; break;};
                    case op_stb:   {strcpy(ops, "stb");     rw=1; break;};
                    case op_stq_u: {strcpy(ops, "stq_u");   rw=1; break;};
                    case op_ldf:   {strcpy(ops, "ldf");     rw=0; break;};
                    case op_ldg:   {strcpy(ops, "ldg");     rw=0; break;};
                    case op_lds:   {strcpy(ops, "lds");     rw=0; break;};
                    case op_ldt:   {strcpy(ops, "ldt");     rw=0; break;};
                    case op_stf:   {strcpy(ops, "stf");     rw=1; break;};
                    case op_stg:   {strcpy(ops, "stg");     rw=1; break;};
                    case op_sts:   {strcpy(ops, "sts");     rw=1; break;};
                    case op_stt:   {strcpy(ops, "stt");     rw=1; break;};
                    case op_ldl:   {strcpy(ops, "ldl");     rw=0; break;};
                    case op_ldq:   {strcpy(ops, "ldq");     rw=0; break;};
                    case op_ldl_l: {strcpy(ops, "ldl_l");   rw=0; break;};
                    case op_ldq_l: {strcpy(ops, "ldq_l");   rw=0; break;};
                    case op_stl:   {strcpy(ops, "stl");     rw=1; break;};
                    case op_stq:   {strcpy(ops, "stq");     rw=1; break;};
                    case op_stl_c: {strcpy(ops, "stl_c");   rw=1; break;};
                    case op_stq_c: {strcpy(ops, "stq_c");   rw=1; break;};
    
                    default:       {strcpy(ops, "unknown"); rw=0; break;};
                  }

                  if (irb<29)
                  {                 
                    if (pass==0)
                    {
                      AddCallInst(i, InstBefore, "MPCSaveAddr", EffAddrValue, proc);
                      AddCallInst(i, InstBefore, "MPCRegAccess", proc);
                    }
                    else if (pass>0)
                    {
                      AddCallInst(i, InstBefore, "MPCBigRegAccess", proc, EffAddrValue, xaddr1);
                    }
                    else if (pass<0)
                    {
                      AddCallInst(i, InstBefore, "MPCBigRegAccess", 0, EffAddrValue, xaddr1);
                    }

                    fprintf(fmpc5,"inst={%5u}, type={%6s}, reg={%3u}, offset={%6d})\n", 
                                   inst_max, ops, irb, imd);

                    inst_max++;
                  }
                }
              }
            }
          }

          if (inst_max>inst_max1) inst_max1=inst_max;

          fclose(fmpc5);
        }
      }  

      proc++;
    }
  } while (ret!=EOF);

  /************************************************************/
  printf("\nEbedding additional code ...\n");

  if (pass>=0)
    AddCallProgram(ProgramBefore,"MPCInit", proc, pass, xaddr2-xaddr1);
  else if (pass<0)
    AddCallProgram(ProgramBefore,"MPCInit", 1, 999999999, xaddr2-xaddr1);

  AddCallProgram(ProgramAfter,"MPCFinish");

  printf("\nInstrumentation finished!\n");

  fclose(fmpc2);
  fclose(fmpc1);
}
