/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _tags_h_
#define _tags_h_

#include "Objects.h"

// Useful macro definitions

#define TYPE_INDEX(_t_)	     ((_t_) - T_INT)
#define IS_BIN_OP(_e_)	     ((_e_) >= EQ_OP && (_e_) <= NEQV_OP)
#define IS_ARITH_TYPE(_t_)   ((_t_) >= T_INT && (_t_) <= T_BOOL)

       /////////////////////
       //
       //  Tag values identifying Statement categories
       //
typedef enum statement_tag {
   GLOBAL = 100,
   PROG_HEDR = 101,
   PROC_HEDR = 102,
   BASIC_BLOCK = 103,
   CONTROL_END = 104,
   IF_NODE = 105,
   LOOP_NODE = 106,
   FOR_NODE = 107,
   FORALL_NODE = 108,
   WHILE_NODE = 109,
   EXIT_NODE = 110,
   ASSIGN_STAT = 111,
   M_ASSIGN_STAT = 112,
   PROC_STAT = 113,
   SUM_ACC = 114,
   MULT_ACC = 115,
   MAX_ACC = 116,
   MIN_ACC = 117,
   CAT_ACC = 118,
   OR_ACC = 119,
   AND_ACC = 120,
   READ_STAT = 121,
   WRITE_STAT = 122,
   OTHERIO_STAT = 123,
   CDOALL_NODE = 124,
   SDOALL_NODE = 125,
   DOACROSS_NODE = 126,
   CDOACROSS_NODE = 127,
   FUNC_HEDR = 130,
   WHERE_NODE = 131,
   ALLDO_NODE = 132,
   IDENTIFY = 133,
   FORMAT_STAT = 134,
   STOP_STAT = 135,
   RETURN_STAT = 136,
   ELSEIF_NODE = 137,
   ARITHIF_NODE = 138,
   GOTO_NODE = 139,
   ASSGOTO_NODE = 140,
   COMGOTO_NODE = 141,
   PAUSE_NODE = 142,
   STOP_NODE = 143,
   LOGIF_NODE = 145,
   ASSLAB_STAT = 146,
   BLOB = 150, 
   SIZES = 151,
   COMMENT_STAT = 152,
   CONT_STAT = 153,
   VAR_DECL = 154,
   PARAM_DECL = 155,
   COMM_STAT = 156,
   EQUI_STAT = 157,
   IMPL_DECL = 158,
   DATA_DECL = 159,
   SAVE_DECL = 160,
   ENTRY_STAT = 162,
   STMTFN_STAT = 163,
   DIM_STAT = 164,
   BLOCK_DATA = 165,
   EXTERN_STAT = 166,
   INTRIN_STAT = 167,
   ENUM_DECL = 168,
   CLASS_DECL = 169,
   UNION_DECL = 170,
   STRUCT_DECL = 171,
   DERIVED_CLASS_DECL = 172,
   EXPR_STMT_NODE = 173,
   DO_WHILE_NODE = 174,
   SWITCH_NODE = 175,
   CASE_NODE = 176,
   DEFAULT_NODE = 177,
   BREAK_NODE = 178,
   CONTINUE_NODE = 179,
   RETURN_NODE = 180,
   ASM_NODE = 181,
   COBREAK_NODE = 182,
   COLOOP_NODE = 183,
   COEXEC_NODE = 184,
   LABEL_STAT = 185,
   TASK_COM = 186,
   ATTR_DECL = 187,
   NAMELIST_STAT = 188,
   FUTURE_STMT = 189,
   COLLECTION_DECL = 190,
   PRINT_STAT = 200,
   BACKSPACE_STAT = 201,
   REWIND_STAT = 202,
   ENDFILE_STAT = 203,
   INQUIRE_STAT = 204,
   OPEN_STAT = 205,
   CLOSE_STAT = 206,
   INCLUDE_STAT = 208,
   PARDO_NODE = 225,
   PARSECTIONS_NODE = 226,
   SECTION_NODE = 227,
   GUARDS_NODE = 228,
   LOCK_NODE = 229,
   UNLOCK_NODE = 230,
   CRITSECTION_NODE = 231,
   POST_NODE = 232,
   WAIT_NODE = 233,
   CLEAR_NODE = 234,
   POSTSEQ_NODE = 235,
   WAITSEQ_NODE = 236,
   SETSEQ_NODE = 237,
   ASSIGN_NODE = 238,
   RELEASE_NODE = 239,
   PRIVATE_NODE = 240,
   SCOMMON_NODE = 241,
   PARREGION_NODE = 242,
   PDO_NODE = 243,
   PSECTIONS_NODE = 244,
   SINGLEPROCESS_NODE = 245,
   SKIPPASTEOF_NODE = 246,
   INTENT_STMT = 250,
   OPTIONAL_STMT = 251,
   PUBLIC_STMT = 252,
   PRIVATE_STMT = 253,
   ALLOCATABLE_STMT = 254,
   POINTER_STMT = 255,
   TARGET_STMT = 256,
   ALLOCATE_STMT = 257,
   NULLIFY_STMT = 258,
   DEALLOCATE_STMT = 259,
   SEQUENCE_STMT  = 260,
   CYCLE_STMT = 261,
   EXIT_STMT = 262,
   CONTAINS_STMT = 263,
   WHERE_BLOCK_STMT = 264,
   MODULE_STMT = 265,
   USE_STMT = 266,
   INTERFACE_STMT = 267,
   MODULE_PROC_STMT = 268,
   OVERLOADED_ASSIGN_STAT = 269,
   POINTER_ASSIGN_STAT = 270,
   OVERLOADED_PROC_STAT = 271,
   DECOMPOSITION_STMT = 275,
   ALIGN_STMT = 276,
   DISTRIBUTE_STMT = 277,
   REDUCE_STMT = 278,
   CVX_ACCEPT_STMT = 700, // included for Convex fortran directives
   CVX_DECODE_STMT = 701,
   CVX_ENCODE_STMT = 702,
   CVX_FIND_STMT = 703,
   CVX_OPTIONS_STMT = 704,
   CVX_TYPE_STMT = 705,
   CVX_DIRECTIVE = 706,
   CRI_DIRECTIVE = 780,   // included for CRI F90 directives
   CRI_AUTOTASKING = 781
 } statement_tag;

       /////////////////////
       //
       //  Tag values identifying Expression categories
       //
typedef enum expression_tag {
   DEFAULT_EXPR = 299,
   INT_VAL = 300,
   FLOAT_VAL = 301,
   DOUBLE_VAL = 302,
   BOOL_VAL = 303,
   CHAR_VAL = 304,
   STRING_VAL = 305,
   KEYWORD_VAL = 319,
   COMPLEX_VAL = 317,
   OCTAL_VAL = 480,
   HEX_VAL = 481,
   CONST_REF = 306,
   VAR_REF = 307,
   ARRAY_REF = 308,
   RECORD_REF = 309,
   ENUM_REF = 310,
   LABEL_REF = 318,
   VAR_LIST = 311,
   EXPR_LIST = 312,
   RANGE_LIST = 313,
   CASE_CHOICE = 314,
   DEF_CHOICE = 315,
   VARIANT_CHOICE = 316,
   DDOT = 324,
   RANGE_OP = 325,
   UPPER_OP = 326,
   LOWER_OP = 327,
   EQ_OP = 328,
   LT_OP = 329,
   GT_OP = 330,
   NOTEQL_OP = 331,
   LTEQL_OP = 332,
   GTEQL_OP = 333,
   ADD_OP = 334,
   SUBT_OP = 335,
   OR_OP = 336,
   MULT_OP = 337,
   DIV_OP = 338,
   MOD_OP = 339,
   AND_OP = 340,
   EXP_OP = 341,
   ARRAY_MULT = 342,
   CONCAT_OP = 343,
   XOR_OP = 344,
   EQV_OP = 345,
   NEQV_OP = 346,
   MINUS_OP = 350,
   NOT_OP = 351,
   ASSGN_OP = 352,
   DEREF_OP = 353,
   POINTST_OP = 354,
   FUNCTION_OP = 355,
   MINUSMINUS_OP = 356,
   PLUSPLUS_OP = 357,
   BITAND_OP = 358,
   BITOR_OP = 359,
   STAR_RANGE = 360,
   PROC_CALL = 370,
   FUNC_CALL = 371,
   CONSTRUCTOR_REF = 380,
   ACCESS_REF = 381,
   CONS = 382,
   ACCESS = 383,
   IOACCESS = 384,
   CONTROL_LIST = 385,
   SEQ = 386,
   SPEC_PAIR = 387,
   COMM_LIST = 388,
   STMT_STR = 389,
   EQUI_LIST = 390,
   IMPL_TYPE = 391,
   STMTFN_DECL = 392,
   BIT_COMPLEMENT_OP = 393,
   EXPR_IF = 394,
   EXPR_IF_BODY = 395,
   FUNCTION_REF = 396,
   LSHIFT_OP = 397,
   RSHIFT_OP = 398,
   UNARY_ADD_OP = 399,
   SIZE_OP = 400,
   INTEGER_DIV_OP = 401,
   SUB_OP = 402,
   LE_OP = 403,
   GE_OP = 404,
   NE_OP = 405,
   CLASSINIT_OP = 406,
   CAST_OP = 407,
   ADDRESS_OP = 408,
   POINTSAT_OP = 409,
   COPY_NODE = 410,
   INIT_LIST = 411,
   VECTOR_CONST = 412,
   BIT_NUMBER = 413,
   ARITH_ASSGN_OP = 414,
   ARRAY_OP = 415,
   NEW_OP = 416,
   DELETE_OP = 417,
   NAMELIST_LIST = 418,
   THIS_NODE = 419,
   SCOPE_OP = 420,
   PLUS_ASSGN_OP = 421,
   MINUS_ASSGN_OP = 422,
   AND_ASSGN_OP = 423,
   IOR_ASSGN_OP = 424,
   MULT_ASSGN_OP = 425,
   DIV_ASSGN_OP = 426,
   MOD_ASSGN_OP = 427,
   XOR_ASSGN_OP = 428,
   LSHIFT_ASSGN_OP = 429,
   RSHIFT_ASSGN_OP = 430,
   ORDERED_OP = 431,
   EXTEND_OP = 432,
   MAXPARALLEL_OP = 433,
   SAMETYPE_OP = 434,
   TYPE_REF = 450,
   STRUCTURE_CONSTRUCTOR = 451,
   ARRAY_CONSTRUCTOR = 452,
   SECTION_REF = 453,
   VECTOR_SUBSCRIPT = 454,
   SECTION_OPERANDS = 455,
   KEYWORD_ARG = 456,
   OVERLOADED_CALL = 457,
   INTERFACE_REF = 458,
   RENAME_NODE = 459,
   TYPE_NODE = 460,
   PAREN_OP = 461,
   PARAMETER_OP = 462,
   PUBLIC_OP = 463,
   PRIVATE_OP = 464,
   ALLOCATABLE_OP = 465,
   DIMENSION_OP = 466,
   EXTERNAL_OP = 467,
   IN_OP = 468,
   OUT_OP = 469,
   INOUT_OP = 470,
   INTRINSIC_OP = 471,
   POINTER_OP = 472,
   OPTIONAL_OP = 473,
   SAVE_OP = 474,
   TARGET_OP = 475,
   ONLY_NODE = 476,
   LEN_OP = 477,
   TYPE_OP = 479,
   ASSIGN_LOCK_DIR = 707, // added for Convex directives
   FREE_LOCK_DIR = 708,
   BEGIN_ORDER_DIR = 709,
   BEGIN_TASKS_DIR = 710,
   END_TASKS_DIR = 711,
   BEGIN_SECTION_DIR = 712,
   END_SECTION_DIR = 713,
   FORCE_PARALLEL_EXT_DIR = 714,
   FORCE_PARALLEL_DIR = 715,
   FORCE_VECTOR_DIR = 716,
   MAX_TRIPS_DIR = 718,
   NEXT_TASK_DIR = 719,
   NO_PARALLEL_DIR = 720,
   NO_PEEL_DIR = 721,
   NO_PROMOTE_TEST_DIR = 723,
   NO_RECURRENCE_DIR = 724,
   NO_SIDE_EFFECTS_DIR = 725,
   NO_VECTOR_DIR = 726,
   PEEL_ALL_DIR = 727,
   PEEL_DIR = 728,
   PREFER_PARALLEL_EXT_DIR = 731,
   PREFER_PARALLEL_DIR = 732,
   PREFER_VECTOR_DIR = 733,
   PROMOTE_TEST_DIR = 729,
   PROMOTE_TEST_ALL_DIR = 730,
   PSTRIP_DIR = 734,
   ROW_WISE_DIR = 735,
   SCALAR_DIR = 736,
   SELECT_DIR = 737,
   SYNCH_PARALLEL_DIR = 738,
   UNROLL_DIR = 739,
   VSTRIP_DIR = 740,
   END_ORDER_DIR = 741,
   CRI_GUARD_DIR = 742,          // added for CRI directives
   CRI_ENDGUARD_DIR = 743,
   CRI_PARALLEL_DIR = 744,
   CRI_CASE_DIR = 745,
   CRI_ENDCASE_DIR = 746,
   CRI_ENDPARALLEL_DIR = 747,
   CRI_DOALLAUTO_DIR = 748,
   CRI_IVDEP_DIR = 749,
   CRI_NOTASK_DIR = 750,
   CRI_TASK_DIR = 751,
   CRI_NOSF_DIR = 752,
   CRI_NOVECTOR_DIR = 753,
   CRI_VECTOR_DIR = 754,
   CRI_PREFERTASK_DIR = 755,
   CRI_PREFERVECTOR_DIR = 756,
   CRI_CHUNKSIZE_DIR = 757,
   CRI_NEXTSCALAR_DIR = 758,
   CRI_UNROLL_DIR = 759,
   IMPL_CONV = 800,              // added for C++ implicit conversions
   INIT_DECLARATOR = 801,        // added for C++ initialised declarators
   DECLARATION = 802,            // added for C++ declarations
   POINTER_MEMBER_FUN = 803,     // added for C++ ->* operator
   SELECT_MEMBER_FUN = 804,      // added for C++ .* operator
   FIELD_SELECT_OP = 805,        // added for C++ (overloadable) . operator
   NEW_ARRAY_OP = 806,           // added for C++ 'new []' operator
   DELETE_ARRAY_OP = 806         // added for C++ 'delete []' operator
 } expression_tag;

       /////////////////////
       //
       //  Tag values identifying Symbol categories
       //
typedef enum symbol_tag {
   DEFAULT_NAME = 449,
   CONST_NAME = 500,
   ENUM_NAME = 501,
   FIELD_NAME = 502,
   VARIABLE_NAME = 503,
   TYPE_NAME = 504,
   PROGRAM_NAME = 505,
   PROCEDURE_NAME = 506,
   VAR_FIELD = 507,
   LABEL_VAR = 508,
   FUNCTION_NAME = 509,
   MEMBER_FUNC = 510,
   CLASS_NAME = 511,
   UNION_NAME = 512,
   STRUCT_NAME = 513,
   LABEL_NAME = 514,
   COLLECTION_NAME = 515,
   ROUTINE_NAME = 516,
   CONSTRUCT_NAME = 517,
   INTERFACE_NAME = 518,
   MODULE_NAME = 519,
   TEMPLATE_NAME = 520,
   NAMESPACE_NAME = 521,
   ANON_FIELD_NAME = 522,
 } symbol_tag;

       /////////////////////
       //
       //  Tag values identifying Type categories
       //
typedef enum type_tag {
   T_DEFAULT = 550,
   T_INT = 551,
   T_FLOAT = 552,
   T_DOUBLE = 553,
   T_CHAR = 554,
   T_BOOL = 555,
   T_STRING = 556,
   T_COMPLEX = 564,
   T_ENUM = 557,
   T_SUBRANGE = 558,
   T_LIST = 559,
   T_ARRAY = 560,
   T_RECORD = 561,
   T_ENUM_FIELD = 562,
   T_UNKNOWN = 563,
   T_VOID = 565,
   T_DESCRIPT = 566,
   T_FUNCTION = 567,
   T_POINTER = 568,
   T_UNION = 569,
   T_STRUCT = 570,
   T_CLASS = 571,
   T_DERIVED_CLASS = 572,
   T_DERIVED_TYPE = 573,
   T_GATE = 580,
   T_EVENT = 581,
   T_SEQUENCE = 582,
   T_DCOMPLEX = 583,
   T_WCHART = 584
 } type_tag;

       /////////////////////
       //
       //  Tags to differentiate variable usages
       //
typedef enum variable_tag {
   LOCAL = 600,
   INPUT = 601,
   OUTPUT = 602,
   IO = 603
 } variable_tag;

/* Enumerator type for declaration status.
 *  These are combined as they are seen during the parse and keep track
 *  of what qualifiers and type specifiers have been recognised.
 */

       /////////////////////
       //
       // Type flags used by C and C++ parsers. These are combined
       // in boolean bit-vectors to indicate the type specifiers
       // that apply to particular type node.
       //
typedef enum {
  NO_SPECS = 0x0,	 /* have not seen any decl_specs	*/

  GOT_CONST = 0x01,	 /* have seen a 'const' specifier	*/
  GOT_VOLATILE = 0x02,	 /* have seen a 'volatile' specifier	*/
  GOT_MUTABLE = 0x04,	 /* have seen a 'mutable' specifier	*/

  GOT_AUTO = 0x08,	 /* have seen an 'auto' specifier	*/
  GOT_REGISTER = 0x10,	 /* have seen a 'register' specifier	*/
  GOT_STATIC = 0x20,	 /* have seen a 'static' specifier	*/
  GOT_EXTERN = 0x40,	 /* have seen an 'extern' specifier	*/

  GOT_INLINE = 0x80,	 /* have seen an 'inline' specifier	*/
  GOT_FRIEND = 0x100,	 /* have seen a 'friend' specifier	*/
  GOT_VIRTUAL = 0x200,	 /* have seen a 'virtual' specifier	*/
  GOT_TYPEDEF = 0x400,	 /* have seen a 'typedef' specifier	*/
  GOT_EXPLICIT = 0x800,  /* have seen an 'explicit' specifier	*/

  GOT_TYPENAME = 0x1000, /* have seen a 'type_name'		*/
  GOT_CLASSNAME = 0x2000,/* have seen a 'class_name'		*/
  GOT_ENUMNAME = 0x4000, /* have seen an 'enum_name' specifier	*/

  GOT_INT = 0x8000,      /* have seen an 'int' specifier	*/
  GOT_CHAR = 0x10000,	 /* have seen a 'char' specifier	*/
  GOT_FLOAT = 0x20000,   /* have seen a 'float' specifier	*/
  GOT_DOUBLE = 0x40000,  /* have seen a 'double' specifier	*/
  GOT_BOOL = 0x80000,    /* have seen a 'bool' specifier	*/
  GOT_VOID = 0x100000,	 /* have seen a 'void' specifier	*/

  GOT_SIGNED = 0x200000, /* have seen a 'signed' specifier	*/
  GOT_UNSIGNED = 0x400000,/*have seen an 'unsigned' specifier	*/

  GOT_LONG = 0x800000,	 /* have seen a 'long' specifier	*/
  GOT_SHORT = 0x1000000, /* have seen a 'short'specifier	*/

  GOT_ERROR = 0x2000000, /* have encountered type error         */
  FINALISED = 0x4000000  /* declaration specifier is finalised  */
} TypeSpecifier ;

#define ANY_TYPE (GOT_INT + GOT_CHAR + GOT_FLOAT + GOT_DOUBLE + GOT_BOOL + GOT_VOID)
#define ANY_SIGN (GOT_SIGNED + GOT_UNSIGNED)
#define ANY_SIZE (GOT_LONG + GOT_SHORT)
#define ANY_MODS (GOT_AUTO + GOT_REGISTER + GOT_STATIC + GOT_EXTERN)
#define ANY_STORAGE_CLASS (GOT_AUTO + GOT_REGISTER + GOT_STATIC + GOT_EXTERN + GOT_MUTABLE)

#define HAS_TYPE(_m_) ((_m_ & ANY_TYPE) != 0)
#define HAS_SIGN(_m_) ((_m_ & ANY_SIGN) != 0)
#define HAS_SIZE(_m_) ((_m_ & ANY_SIZE) != 0)
#define HAS_MODS(_m_) ((_m_ & ANY_MODS) != 0)
#define HAS_STORAGE_CLASS(_m_) ((_m_ & ANY_STORAGE_CLASS) != 0)

#define TYPE_SHIFT 15
#define SIGN_SHIFT 21
#define SIZE_SHIFT 23
#define MODS_SHIFT 0
#define STOR_SHIFT 3

#define INT_FLAG (GOT_INT >> TYPE_SHIFT)
#define CHAR_FLAG (GOT_CHAR >> TYPE_SHIFT)
#define FLOAT_FLAG (GOT_FLOAT >> TYPE_SHIFT)
#define DOUBLE_FLAG (GOT_DOUBLE >> TYPE_SHIFT)
#define BOOL_FLAG (GOT_BOOL >> TYPE_SHIFT)
#define VOID_FLAG (GOT_VOID >> TYPE_SHIFT)

#define SIGNED_FLAG (GOT_SIGNED >> SIGN_SHIFT)
#define UNSIGNED_FLAG (GOT_UNSIGNED >> SIGN_SHIFT)

#define LONG_FLAG (GOT_LONG >> SIZE_SHIFT)
#define SHORT_FLAG (GOT_SHORT >> SIZE_SHIFT)

#define AUTO_FLAG (GOT_AUTO >> STOR_SHIFT)
#define STATIC_FLAG (GOT_STATIC >> STOR_SHIFT)
#define REGISTER_FLAG (GOT_REGISTER >> STOR_SHIFT)
#define EXTERN_FLAG (GOT_EXTERN >> STOR_SHIFT)

/*--------------------------------------------------------------

CLASS
    Tag

    Union of all enumerated tag types - used to tag parse-tree entities.

DESCRIPTION

    Every parse-tree entity (Statement, Symbol, Expression, Label,
    Type) is publicy-derived from the Tag class. This means that
    each such object is tagged to indicate, for example what type of
    Statement it is, or what Expression operator is represent.

    A variety of predicates on Tag objects are provided, and these
    are of course inherited by the Symbol, Statements etc. which
    are derived from the Tag class.

    All classes derived publicly from the Tag class are also
    a publicly derived sub-class of Enumerable objects. This means
    those objects can be contained in Map, SetOf, SymbolTable, and other
    container classes constructed using indexed List objects as
    their low-level container.

    Note the Tag constructor does not initialise the index of
    the Enumerable object from which it it derived. This must be
    done in the constructor at the next level up in the class 
    hierarchy by calling the Enumerable::set_index() method.
---------------------------------------------------------------*/

class Tag : public Enumerable {

private:

  union {
    statement_tag  stat_tag ;
    expression_tag expr_tag ;
    symbol_tag     symb_tag ;
    type_tag       typ_tag ;
    variable_tag   var_tag ;
    int            value ;
  } tag ;

public:

       /////////////////////
       //
       // Actionless default constructor
       //
  Tag ()                          { }

       /////////////////////
       //
       //
  Tag (int v)                     { tag.value = v ;       }

       /////////////////////
       //
       //
  statement_tag stmt_variant ()   { return tag.stat_tag ; }

       /////////////////////
       //
       //
  expression_tag expr_variant ()  { return tag.expr_tag ; }

       /////////////////////
       //
       //
  symbol_tag symb_variant ()      { return tag.symb_tag ; }

       /////////////////////
       //
       //
  type_tag type_variant ()        { return tag.typ_tag ;  }

       /////////////////////
       //
       //
  variable_tag var_variant ()     { return tag.var_tag ;  }

       /////////////////////
       //
       //
  int variant ()                  { return tag.value ;    }

  // following functions ought to range-check their parameter
  // ...implement this later.

       /////////////////////
       //
       //
  void stmt_variant (statement_tag v)  { tag.stat_tag = v ; }

       /////////////////////
       //
       //
  void expr_variant (expression_tag v) { tag.expr_tag = v ; }

       /////////////////////
       //
       //
  void symb_variant (symbol_tag v)     { tag.symb_tag = v ; }

       /////////////////////
       //
       //
  void type_variant (type_tag v)       { tag.typ_tag  = v ; }

       /////////////////////
       //
       //
  void var_variant (variable_tag v)    { tag.var_tag  = v ; }

       /////////////////////
       //
       //
  void variant (int v)                 { tag.value    = v ; }
  
       /////////////////////
       //
       //
  int has_scope ()
    { return (
        (tag.stat_tag == FORALL_NODE)|| 
        (tag.stat_tag == CDOALL_NODE)|| 
        (tag.stat_tag == SDOALL_NODE)|| 
        (tag.stat_tag == DOACROSS_NODE)|| 
        (tag.stat_tag == CDOACROSS_NODE)||
        (tag.stat_tag == FOR_NODE)
        ) ;
    }

  // predicates for statement and expression tags

       /////////////////////
       //
       //
  int is_executable ()
  {
    return 
      !(tag.stat_tag == 104) &&
      !(tag.stat_tag >= 150 && tag.stat_tag <= 152) &&
      !(tag.stat_tag >= 154 && tag.stat_tag <= 172) &&
      !(tag.stat_tag >= 250 && tag.stat_tag <= 271) ;
  }

       /////////////////////
       //
       // Returns true if this stat_tag defines a Statement
       // that might require to know the type of a 
       // variable declared in the same Fortran unit.
  int post_declaration ()
  {
    return is_executable() 
      || (tag.stat_tag == DATA_DECL) 
      || (tag.stat_tag == COMM_STAT) 
      || (tag.stat_tag == EQUI_STAT) 
      || (tag.stat_tag == BLOCK_DATA) 
      || (tag.stat_tag == STOP_STAT)
      || (tag.stat_tag == RETURN_STAT)
      || (tag.stat_tag == CONTROL_END) 
      || (tag.stat_tag == FORMAT_STAT);
  }

  int is_class_name ()
    {
      return (tag.symb_tag == CLASS_NAME) 
	|| (tag.symb_tag == STRUCT_NAME) 
	|| (tag.symb_tag == UNION_NAME) ;
    }    
};

///////////////////////////////////////////////////////
//
// When a user-defined object type is referenced in
// the abstract syntax tree, it may be just a refernce
// or it may be the defining reference.
//
// These two cases are distinguished by the following
// enumerated tags. The semantic values associated with
// struct, union and enum specifiers are tagged with
// one of these tags.
//
typedef enum declaration_tag {
  TYPE_REFERENCE,
  TYPE_DEFINITION
} declaration_tag;

///////////////////////////
//
typedef enum scope_tag {
  CLASS_SCOPE,
  FILE_SCOPE,
  LOCAL_SCOPE,
  FUNCTION_SCOPE,
  NAMESPACE_SCOPE
} scope_tag ;

///////////////////////////
//
// Converts from a symbol_tag to a string representation
// of that symbol tag.
//
// [in] symbol_tag
// [out] const char*
//
const char* enumToText (symbol_tag) ;

#endif // _tags_h_

