/*
 * EOS Timers (Loop or other level) Windows (analysis)
 *
 * Copyright (C) 2000-2004 by Grigori G. Fursin
 *
 * homepage: http://homepages.inf.ed.ac.uk/gfursin
 *
 * ICSA, School of Informatics,
 * University of Edinburgh, UK
 */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include <sys/time.h>
#include <limits.h>

#define timers_max $EOS1$
#define queue_max  $EOS2$

#define MAX_ETIME 86400   

static int    first_run=0;

static int    queue=0;
static long   queue_total=0;

static double Timers[queue_max+1][timers_max+1];
static double Timers1[queue_max+1][timers_max+1];
static long   Calls[queue_max+1][timers_max+1];

/* static variable that holds the initial value of the interval timer */
static struct itimerval first_u; /* user time */
static struct itimerval first_r; /* real time */

/* 
 * elapsed user time routines 
 */

/* init the timer */
static void init_etime(void)
{
    first_u.it_interval.tv_sec = 0;
    first_u.it_interval.tv_usec = 0;
    first_u.it_value.tv_sec = MAX_ETIME;
    first_u.it_value.tv_usec = 0;
    setitimer(ITIMER_VIRTUAL, &first_u, NULL);
    first_r.it_interval.tv_sec = 0;
    first_r.it_interval.tv_usec = 0;
    first_r.it_value.tv_sec = MAX_ETIME;
    first_r.it_value.tv_usec = 0;
    setitimer(ITIMER_REAL, &first_r, NULL);
}

/* return elapsed virtual seconds since call to init_etime */
static double get_etime(void) {
    struct itimerval curr;

    getitimer(ITIMER_VIRTUAL, &curr);
    return (double) ((first_u.it_value.tv_sec - curr.it_value.tv_sec) +
		     (first_u.it_value.tv_usec - curr.it_value.tv_usec)*0.000001);
}

/* return elapsed real seconds since call to init_etime */
static double get_rtime(void) {
    struct itimerval curr;

    getitimer(ITIMER_REAL, &curr);
    return (double) ((first_r.it_value.tv_sec - curr.it_value.tv_sec) +
		     (first_r.it_value.tv_usec - curr.it_value.tv_usec)*0.000001);
}

void eos_timers_init(void)
{
  long y;
  int q;
  FILE *f;

  for (q=0; q<queue_max; q++)
  {
    for (y=1; y<=timers_max; y++)
    {
      Timers[q][y]=0;
      Timers1[q][y]=0;
      Calls[q][y]=0;
    }
  }

  if ((f=fopen("_eos_llprof_out","wt"))==NULL)
  {
    printf("\nFatal Error (FATOM): Can't write file!\n");
    exit(1);
  }
  fclose(f);
}

void save_timers(void)
{
  FILE *f;
  int i=0;

  unsigned int val=0;
  int q1=0;

  double time_all=0;

  if ((f=fopen("_eos_llprof_out","at"))==NULL)
  {
    printf("\nFatal Error (FATOM): Can't write file!\n");
    exit(1);
  }
         
  for (q1=0; q1<queue; q1++)
  {
    fprintf(f, "step={%8u}, ", queue_total);
    queue_total++;

    for (i=1; i<=timers_max; i++)
    {
      fprintf(f, " {%5.2f}", Timers[q1][i]);
      time_all+=Timers[q1][i];

      Timers[q1][i]=0;
      Timers1[q1][i]=0;
      Calls[q1][i]=0;
    }
  
    fprintf(f, "\n");
  }
   
  fclose(f);

  queue=0;
}

void eos_timers_finish(void)
{
  save_timers();
  
  return;
}

void EOS_TIMER1(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  if (first_run==0)
  {
    eos_timers_init();
    atexit(eos_timers_finish);
    queue=0;
    queue_total=0;
    first_run=1;
  }
  
  tmr=*tmr1;

  Calls[queue][tmr]++;

  tt=get_etime();
  Timers1[queue][tmr]=tt;
}

void EOS_TIMER2(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  tmr=*tmr1;
  tt=get_etime();
  Timers[queue][tmr]=tt - Timers1[queue][tmr];

  queue++;

  if (queue>queue_max)
  {
    save_timers();
  }
}

void eos_timer1(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  if (first_run==0)
  {
    eos_timers_init();
    atexit(eos_timers_finish);
    queue=0;
    queue_total=0;
    first_run=1;
  }
  
  tmr=*tmr1;

  Calls[queue][tmr]++;

  tt=get_etime();
  Timers1[queue][tmr]=tt;
}

void eos_timer2(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  tmr=*tmr1;
  tt=get_etime();
  Timers[queue][tmr]=tt - Timers1[queue][tmr];

  queue++;

  if (queue>queue_max)
  {
    save_timers();
  }
}

void eos_timer1_(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  if (first_run==0)
  {
    eos_timers_init();
    atexit(eos_timers_finish);
    queue=0;
    queue_total=0;
    first_run=1;
  }
  
  tmr=*tmr1;

  Calls[queue][tmr]++;

  tt=get_etime();
  Timers1[queue][tmr]=tt;
}

void eos_timer2_(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  tmr=*tmr1;
  tt=get_etime();
  Timers[queue][tmr]=tt - Timers1[queue][tmr];

  queue++;

  if (queue>queue_max)
  {
    save_timers();
  }
}

void eos_timer1__(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  if (first_run==0)
  {
    eos_timers_init();
    atexit(eos_timers_finish);
    queue=0;
    queue_total=0;
    first_run=1;
  }
  
  tmr=*tmr1;

  Calls[queue][tmr]++;

  tt=get_etime();
  Timers1[queue][tmr]=tt;
}

void eos_timer2__(int *tmr1)
{
  int tmr=0;
  double tt=0;
  double tt1=0;

  tmr=*tmr1;
  tt=get_etime();
  Timers[queue][tmr]=tt - Timers1[queue][tmr];

  queue++;

  if (queue>queue_max)
  {
    save_timers();
  }
}

void _Z10eos_timer1i(int tmr)
{
  double tt=0;
  double tt1=0;

  if (first_run==0)
  {
    eos_timers_init();
    atexit(eos_timers_finish);
    queue=0;
    queue_total=0;
    first_run=1;
  }
  
  Calls[queue][tmr]++;

  tt=get_etime();
  Timers1[queue][tmr]=tt;
}

void _Z10eos_timer2i(int tmr)
{
  double tt=0;
  double tt1=0;

  tt=get_etime();
  Timers[queue][tmr]=tt - Timers1[queue][tmr];

  queue++;

  if (queue>queue_max)
  {
    save_timers();
  }
}
