//
// Class CClient for communication with EOS servers
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.net.*;
import java.io.*;

public class CClient
{
  //Variables
  String hostname;
  int port;
  int instrmax;

  String[]       Instr;
  boolean[]      WFE;
  boolean[]      AR;
  static String  err;

  boolean        record=false;
  String         frecord="__instructions.txt";
  BufferedWriter foutr=null;

  int cur;

  //****************************************************************************
  //"Initialization"
  //
  //Input:  hostname1 = Host name
  //        port1     = Port
  //        instrmax1 = Max number of instructions
  //
  //Return: 0  - Ok
  //        >0 - Error
  //

  public int Init(String hostname1,
                  int port1,
                  int instrmax1)
  {
    if (instrmax1<1) return 1;

    hostname=hostname1;
    port=port1;
    instrmax=instrmax1;

    Instr=new String [instrmax];
    WFE  =new boolean [instrmax];
    AR   =new boolean [instrmax];

    cur=0;
    err="";

    return 0;
  }

  //****************************************************************************
  public int Add(String instr, boolean wait_for_echo, boolean add_r)
  {
    if (cur>instrmax) return -1;

    Instr[cur]=instr;
    WFE[cur]=wait_for_echo;
    AR[cur]=add_r;

    cur++;

    return 0;
  }


  //****************************************************************************
  public void RecordInstructions(String str)
  {
    if (str.equals("1")) record=true;
    else record=false;
  }

  //****************************************************************************
  //Return: -5 - if error was returned by client
  //             in transformations it may be ignored
  public int Send()
  {
    int ret=0;

    if (record)
    {
      try
      {
        foutr=new BufferedWriter(new FileWriter(frecord, true));
      }
      catch (IOException e2)
      {
        return -10;
      }

      try
      {
        foutr.write("---------------------------------------------------------------");
        foutr.newLine();
        
        foutr.write("telnet "+hostname+" ");
        foutr.write(String.valueOf(port));
        foutr.newLine();

        foutr.newLine();
      }
      catch (Exception e2)
      {
        try
        {
          foutr.close();
        }
        catch (Exception e3)
        {}

        return -11;
      }

      for (int i=0; i<cur; i++)
      {
        try
        {
          foutr.write(Instr[i]);
          foutr.newLine();
        }
        catch (Exception e2)
        {
          try
          {
            foutr.close();
          }
          catch (Exception e3)
          {}

          return -11;
        }
      }

      try
      {
        foutr.close();
      }
      catch (Exception e5)
      {
      }
    }

    //prepare connection
    Socket sockfd = null;
    DataOutputStream sout = null;
    BufferedReader sin = null;

    try 
    {
      sockfd = new Socket(hostname, port);
      sout = new DataOutputStream(sockfd.getOutputStream());
      sin = new BufferedReader(new InputStreamReader(sockfd.getInputStream()));
    } 
    catch (UnknownHostException e1)
    {
      return -1;
    }
    catch (IOException e2)
    {
      return -2;
    }

    //Sending instructions
    boolean er=false;

    for (int i=0; i<cur; i++)
    {
      ret=send_req(sockfd, sout, sin, Instr[i], WFE[i], AR[i]);
      
      if (ret!=0)
      {
        er=true;
        break;
      }
    }

    cur=0;
  
    try
    {
      sockfd.close();
      sout.close();
      sin.close();
    }
    catch (IOException e) 
    {
//remarked on 2002.10.07
//      if (ret==-2) return -5;
//      if (er) return -3;
//      return -4;
    }

    if (ret==-2) return -5;
    if (er) return -3;

    err="";

    return 0;
  }

  //****************************************************************************
  public String GetLastError()
  {
    String err1=err;
    err="";
    return err1;
  }

  //**************************************************************************
  private static int send_req(Socket sockfd, 
                              DataOutputStream sout, 
                              BufferedReader sin, 
                              String str,
                              boolean wait_for_echo,
                              boolean add_r)
  {
    String buf="";
    String str1=str;
      
    if (add_r) str1+="\r";
    str1+="\n";

    //send
    try
    {
      sout.writeBytes(str1);
    }
    catch (IOException e4)
    {
      err="Error: Can't send data to server!";
      return -1;
    }

    if (!wait_for_echo) return 0;

    //receive echo or response
    try
    {
      buf=sin.readLine();
    }
    catch (IOException e5)
    {
      err="Error: Can't receive echo from remote Server("+e5.getMessage()+")!";
      return -1;
    }

    if (buf==null)
    {
      err="Error: Problems during communicaton with remote Server!";
      return -1;
    }

    if ((buf.length()>=str.length()) &&
        (buf.substring(0, str.length()).toLowerCase().equals(str.toLowerCase())))
    {
      //It was echo. Now recive response
      try
      {
        buf=sin.readLine();
      }
      catch (IOException e5)
      {
        err="Error: Can't receive response from remote Server!";
        return -1;
      }
    }

    if ((buf==null) || (buf.length()<2))
    {
      err="Error: Problems during communicaton with remote Server!";
      return -1;
    }
    
    if (buf.toLowerCase().substring(0,2).equals("ok")) return 0;
    
    err="Error: Received from remote Server: ("+buf+")";
    return -2;
  }

  //**************************************************************************
  private static void wait_for_end(Socket sockfd,
                                   BufferedReader sin)
  {
    String buf="";

    //receive echo or response
    try
    {
      while ((buf=sin.readLine())!=null);
    }
    catch (IOException e5)
    {
    }
  }
}
