//
// Class CLanguage for dealing with different languages
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

public class CLanguages
{
  //Variables
  private CConfig cfg;
  private int numLng;
  private int index;

  private String    f_name, f_ext;
  private String[]  Name;
  private CConfig[] Lng;
  
  //****************************************************************************
  //"Initialization"
  //
  //Input:  cfg1    = configuration object
  //        index1  = index of 'language' in configuration
  //        f_name1 = beginning of the file name with configurations
  //        f_ext1  = file extension
  //
  //Return: 0  - Ok
  //        >0 - Error
  //

  public int Init(CConfig cfg1, int index1, String f_name1, String f_ext1)
  {
    cfg=cfg1;
    index=index1;
    f_name=f_name1;
    f_ext=f_ext1;

    numLng=cfg.GetMaxNumVal(index);;
    if (numLng<1) return 1;

    Name=new String[numLng];
    Lng=new CConfig[numLng];

    return 0;
  }

  //****************************************************************************
  //"Parse Language Configuration Files"
  //
  //Return: <0  - Error
  //        >=0 - Identificator 
  //
  public int Parse()
  {
    int i;

    for (i=0; i<numLng; i++)
    {
      Name[i]=cfg.GetParamStr(index, i);

      String FileCFG=f_name+Name[i]+f_ext;

      //Declaration of class for configuration file parsing
      Lng[i] = new CConfig();

      if (Lng[i].Init(10, 1, FileCFG, 32)!=0) return -1;
      //               |  |        |   |
      //               |  |        |   |---- Max Number of Array Parameters
      //               |  |        | 
      //               |  |        |-------- Name of File with Parameters
      //               |  |
      //               |  |----------------- Max Length of Parameter
      //               |
      //               |-------------------- Max Number of Parameters

      int misc=0;
  
      misc+=Lng[i].Add("t_server_host", "", 1);  //Param 0
      misc+=Lng[i].Add("t_server_port", "", 1);  //Param 1
      misc+=Lng[i].Add("compiler_name", "", 1);  //Param 2
      misc+=Lng[i].Add("compile_only",  "", 1);  //Param 3
      misc+=Lng[i].Add("profile_comp",  "", 1);  //Param 4
      misc+=Lng[i].Add("profile_link",  "", 1);  //Param 5
      misc+=Lng[i].Add("optimization",  "", 9);  //Param 6

      if (misc!=0) return -2;

      misc=Lng[i].Parse();
      if (misc==1) return -3;
      else if (misc==2) return -4;
      else if (misc==3) return -5;
      else if (misc==4) return -6;
      else if (misc==5) return -7;
      else if (misc>0)  return -8;
    }

    return 0;
  }

  //****************************************************************************
  public int GetMaxLanguages()
  {
    return numLng;
  }

  //****************************************************************************
  public String GetLanguageName(int index)
  {
    return Name[index];
  }

  //****************************************************************************
  public int GetLngIndex(String lng)
  {
    int i=0;
    int ret=-1;

    for (i=0; i<numLng; i++)
    {
      if (lng.equals(Name[i]))
      {
        ret=i;
        break;
      }
    }

    return ret;
  }

  //****************************************************************************
  public CConfig GetConfig(int index)
  {
    return Lng[index];
  }
}
