/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#ifndef _comments_h_
#define _comments_h_

#include "Defs.h"
#include "Strings.h"
#include "Objects.h"

// CLASS GROUP: parse tree entity classes

/*--------------------------------------------------------------

CLASS
    Comment

    Contains a single source language comment

DESCRIPTION

 This class is used to create and store, within the parse tree
 a single source comment. The comment itself is stored as
 a String object, together with an integer type field to
 distinguish different types of comments.

 Comments should not be created in isolation, but should be made
 by calling make_comment() from a ParseContext object. This will
 properly associate the newly created comment with that
 ParseContext. 

----------------------------------------------------------------*/

class Comment : public Enumerable {

friend class FortranParser ;

public:

	/////////////////////////////
	//
	// Pointer to the ParseContext containing this comment
	//
  PTR_CNTXT pc ;

	/////////////////////////////
	//
	// Distinguishes different types of comments
	//
  int type;

	/////////////////////////////
	//
  int counter;

	/////////////////////////////
	//
	// The comment string itself
	//
  string cstring ;

	/////////////////////////////
	//
	// Pointer to the next string
	//
  class Comment *next;


  int variant() { return 0 ; } // not used, but g++ needs it (urghh)

	/////////////////////////////
	//
	// Initializes the fields of a comment object
	//
	// [in] ParseContext* the containing context
	// [in] int           the comment type
	// [in] const char*   the comment itself
	//
  void set_fields (PTR_CNTXT c, int ctype, const char *com) {
    pc = c ;
    type = ctype ;
    counter = 0 ;
    next = 0 ;
    cstring = com ;
  }


	/////////////////////////////
	//
	// Initializes the fields of a comment object
	//
	// [in] ParseContext* the containing context
	// [in] int           the comment type
	// [in] char*         the comment itself
	//
  void set_fields (PTR_CNTXT c, int ctype, string *com) {
    pc = c ;
    type = ctype ;
    counter = 0 ;
    next = 0 ;
    cstring = com ;
  }


	/////////////////////////////
	//
	// Constructor.
	//
  Comment (int i) : Enumerable(i), cstring () {}
};

#endif // _comments_h_

