//
// "Edinburgh Optimizing Software" Transformation Server
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

#include "pm.h"
#include "param.h"

#include "global.h"
#include "clparse.h"

#include <iostream.h>

#ifdef W32
#  include <winsock2.h>
#  include <wsipx.h>

   struct ThreadParam {CGlobal* glob; int sock;};

   void ThreadProc (ThreadParam*);
#else
#  include <signal.h>
#  include <stdlib.h>
#  include <sys/wait.h>
#  include <sys/resource.h> //for cpu usage
#  include <unistd.h>

#  include <sys/socket.h>
#  include <sys/types.h>
#  include <memory.h>
#  include <netinet/in.h>

   void sig_chld(int);
   void sig_int(int);
#endif

int proc_num=0;

CGlobal* glob;

int main(int argc, char* argv[])
{
  long misc;
  char FileOpt[]="eos_tr_server.cl";

  //***********************************************************************
  cout.setf(ios::unitbuf);
  cout << "EOS Transformation Server   V1.06   (C) by Grigori G.Fursin   August,2004\n";
  
  //Declaration of class with global variables
  glob = new CGlobal();
    
  //Declaration of class for command-line parsing
  glob->clp = new CCLParse();
     
  if (glob->clp->Init(10, 32, 256, 1, FileOpt)!=0)
  //                   |   |    |  |  | 
  //                   |   |    |  |  |---- Name of File with Options
  //                   |   |    |  |
  //                   |   |    |  |------- Case Sensitive (0) or not (1)
  //                   |   |    |
  //                   |   |    |---------- Max Length of Parameter
  //                   |   |
  //                   |   |--------------- Max Length of Option
  //                   |
  //                   |------------------- Max Number of Options
  {
    cerr << "\nError: Can't initialize object CCLParse!\n";
    return 1;
  }

  misc=0;
  
  misc =glob->clp->Add("--help",      "-h",  "-?", "",     0); //Param 0
  misc+=glob->clp->Add("--version",   "-v",  "",   "",     0); //Param 1
  misc+=glob->clp->Add("--verbose",   "-vb", "",   "",     0); //Param 2
  misc+=glob->clp->Add("--port",      "-p",  "",   "4811", 1); //Param 3
  misc+=glob->clp->Add("--nofork",    "-nf", "",   "",     0); //Param 4
  misc+=glob->clp->Add("--catch",     "-c",  "",   "",     0); //Param 5
  misc+=glob->clp->Add("--terminate", "-tm", "",   "",     0); //Param 6

  if (misc!=0)
  {
    cerr << "\nError: Internal, while initializing CCLParse!\n";
    return 1;
  }

  misc=glob->clp->Parse(argc, argv);
  if (misc==1)
  {
    cerr << "\nError: Unknown option in command line!\n";
    return 1;
  }
  else if (misc==2)
  {
    cerr << "\nError: Option can't be used without parameter in command line!\n";
    return 1;
  }
  else if (misc==3)
  {
    cerr << "\nError: Option can't be used without parameter in file with options!\n";
    return 1;
  }
  else if (misc==4)
  {
    cerr << "\nError: Unknown option in file with options!\n";
    return 1;
  }
  else if (misc>0)
  {
    cerr << "\nError: Internal, while parsing command line!\n";
    return 1;
  }

  if (glob->clp->GetParamLong(paramVersion)==1)
  {
    cout << "\n";
    cout << "Version information:\n";
    cout << "  Version 1.06, released on 2004.August.14\n";

    return 0;
  }

  if (glob->clp->GetParamLong(paramHelp)==1)
  {
    //Help
    cout << "\n";
    cout << "Brief information:\n";    
    cout << "  \"Edinburgh Optimizing Software\" Transformation Server\n";
    cout << "\n";
    cout << "  Copyright (C) 2000-2004 by Grigori G.Fursin\n";
    cout << "  http://homepages.inf.ed.ac.uk/gfursin\n";
    cout << "\n";
    cout << "  ICSA, School of Informatics,\n";
    cout << "  University of Edinburgh, UK\n";
    cout << "\n";
    cout << "Options (you can also add options into file " << FileOpt << "):\n";
    cout << "  --help,      -h, -?            Help;\n";
    cout << "  --version,   -v                Print version;\n";
    cout << "  --verbose,   -vb               Print additional diagnostic messages;\n";
    cout << "  --port,      -p <port number>  Port number for server [Default=4811];\n";
    cout << "  --nofork,    -nf               No fork for child process;\n";
    cout << "  --catch,     -c                Catch signals from child process.\n";
    cout << "  --terminate, -tm               Terminate server after each session.\n";

    return 0;
  }

  if (glob->clp->GetParamLong(paramVerbose)==1)
  {
    //Help
    cout << "\n";
    cout << "Options:\n";
    cout << "    Verbose=" << glob->clp->GetParamLong(paramVerbose) << "\n";
    cout << "       Port=" << glob->clp->GetParamLong(paramPort) << "\n";
    cout << "     NoFork=" << glob->clp->GetParamLong(paramNofork) << "\n";
    cout << "      Catch=" << glob->clp->GetParamLong(paramCatch) << "\n";
    cout << "  Terminate=" << glob->clp->GetParamLong(paramTerminate) << "\n";
  }

#ifdef W32
  //Preparing Server

  //Initialize with the WINSOCK library 
  WORD wVersionRequested=MAKEWORD(1,1);
  WSADATA wsdata;
  
  int rc = WSAStartup(wVersionRequested, &wsdata);

  if (rc) 
  {
    cerr << "\nError: Socket Library Initialization failed!\n";
    return 1;
  }

  SOCKET sockfd, newsockfd;
  char reuse=1;
#else
  //Catching Signals
  signal(SIGCHLD, sig_chld);
  signal(SIGINT, sig_int);

  //Preparing Server
  int sockfd=0, newsockfd=0;
  int reuse=1;
#endif
  
  if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
  {
    cerr << "\nError: Can't open stream socket!\n";
    return 1;
  }

  if (setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse) < 0)
  {
    cerr << "\nError: Can't set REUSE option for socket!\n";
    return 1;
  }

  struct sockaddr_in cli_addr, serv_addr;
  memset (&serv_addr, 0, sizeof(serv_addr));

  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(glob->clp->GetParamInt(paramPort));

  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
  {
    cerr << "\n";
    cerr << "Error: Can't bind host with socket!\n";
    cerr << "       Probably address is already in use!\n";
    return 1;
  }

  if (listen (sockfd, 5)!=0)
  {
    cerr << "\nError: Can't start listening!\n";
    return 1;
  }

  //Start listening
  cout << "\n";  
  cout << "Server is waiting for connections [port "
       << glob->clp->GetParamLong(paramPort)
       << "] ...\n";

  cout << "\n";

  for (;;)
  {
#ifdef W32
    int clilen=sizeof(cli_addr);
#else
    socklen_t clilen=sizeof(cli_addr);
#endif

#ifndef W32
    signal(SIGCHLD, sig_chld); //Since there are some problems 
                               //with termination of child processes...
#endif

    newsockfd=accept (sockfd, (struct sockaddr *)&cli_addr, &clilen);
    if (newsockfd < 0)
    {
//      cerr << "Error: Accept error!\n";  
//      return 1; // don't stop - may be temporal error
      continue;
    }

    if (glob->clp->GetParamLong(paramNofork)==0)
    {
      //If Fork
      proc_num++;
      cout << "[" << proc_num << "] Child process activated ...\n";

#ifdef W32
      ThreadParam tp1;
      ThreadParam* tp=&tp1;

      tp->sock=newsockfd;
      tp->glob=glob;
      unsigned long ThreadId;

      if (!CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ThreadProc,
                       (ThreadParam*) tp, 0, & ThreadId ))
      {
        cerr << "\nError: Can't create new thread!\n";
//      return 1; // don't stop - may be temporal error
        continue;
      }
#else
      int childpid;
      if ((childpid = fork()) < 0)
      {
        cerr << "\nError: Fork error!\n";
//        return 1; // don't stop - may be temporal error
        continue;
      }
      
      if (childpid == 0)
      {
        // Child process - talk with connected client
        ProcessMessages (newsockfd, glob);
        if (shutdown (newsockfd, 2)!=0)
        {
          cerr << "Error: Can't close socket connection!\n";
        }

        close(newsockfd);
        return 0;
      }
      else
      {
        //Parent
        //Continue listening (socket will be closed by child process)

        close(newsockfd);
      }
#endif
    }
    else
    {
      // If No Fork
      cout << "Connection established ...\n";

      ProcessMessages (newsockfd, glob);
      if (shutdown (newsockfd, 2)!=0)
      {
        cerr << "Error: Can't close socket connection!\n";
      }

#ifdef W32
      closesocket(newsockfd);
#else
      close(newsockfd);
#endif

      cout << "Connection closed ...\n";

      if (glob->clp->GetParamLong(paramTerminate)==1) exit(0);
    }
  }

//  cout << "\nProgram finished successfully!\n";
//  return 0;
}

#ifdef W32
void ThreadProc (ThreadParam* tp)
{
  // Child process - talk with connected client
  int sockfd=tp->sock;
  CGlobal* g=tp->glob;

  ProcessMessages (sockfd, g);
  if (shutdown (sockfd, 2)!=0)
  {
    cerr << "Error: Can't close socket connection!\n";
  }

  closesocket(sockfd);

  proc_num--;
  if (proc_num>=0)
  {
    cout << "[" << proc_num << "] Child process terminated ...\n";
  }

  if (glob->clp->GetParamLong(paramTerminate)==1) exit(0);

  return;
}
#else
void sig_int(int signo)
{
  double user, sys;
  struct rusage	myusage, childusage;

  cout << "\nCtrl+C detected. Terminating ...\n";
  
  if (getrusage(RUSAGE_SELF, &myusage) < 0)
  {
    cerr << "\nError: Can't use function 'getrusage'\n";
    exit(1);  
  }
  if (getrusage(RUSAGE_CHILDREN, &childusage) < 0)
  {
    cerr << "\nError: Can't use function 'getrusage'\n";
    exit(1);  
  }
  
  user = (double) myusage.ru_utime.tv_sec +
         myusage.ru_utime.tv_usec/1000000.0;
  user += (double) childusage.ru_utime.tv_sec +
          childusage.ru_utime.tv_usec/1000000.0;
  sys = (double) myusage.ru_stime.tv_sec +
        myusage.ru_stime.tv_usec/1000000.0;
  sys += (double) childusage.ru_stime.tv_sec +
         childusage.ru_stime.tv_usec/1000000.0;
  
  cout << "\n";
  cout << "  User Time = " << user << " s. \n";
  cout << "System Time = " << sys  << " s. \n";

  cout << "\nProgram finished successfully!\n";
  exit(0);
}

void sig_chld(int signo)
{
  pid_t	pid;
  int stat;
  
  while ( (pid = waitpid(-1, &stat, WNOHANG)) > 0) {}
  proc_num--;
  if (proc_num>=0)
  {
    cout << "[" << proc_num << "] Child process terminated ...\n";
  }

  return;
}
#endif
